# Use at your own risk.

import os
import datetime
import random

# Global variables - mess with these values but stay sensible, don't put high range parameters lower than low range parameters,
# for example as there is no handling code for it.

# amount of patches to be generated.
numOfPatches = 25

# How many events within patch
minNumOfEvents = 24
maxNumOfEvents = 32 
# The maximum allowed in Pattern Mutator is 340, if using the doubling the max can be no more than 170.

# Event Globals - Don't remove any of the +1 offsets
noteStartMin = 1
noteStartMax = 32
noteRangeLow = 24 + 1 # lowest note that can be randomly generated
noteRangeHigh = 84 + 1 # highest note that can be randomly generated
noteVelLow = 50 + 1
noteVelHigh = 110 + 1
noteLengthLow = 1 + 1
noteLengthHigh = 4 + 1
noteLSFLow = 50 + 1
noteLSFHigh = 200 + 1

# In this example, CV output params represent a gentle filter setup, where CV1 is the freq and CV2 is the res.
# The range which must be used is from 1-255, where 1 is a parameter value of 0, and the 255 will scale to a parameter value of 127 on
# a standard Reason filter. If you wish to use a bipolar connection, then 1 is -127 and 255 is 127, with 128 being your 0-crossing.
# This isn't the whole story with CV, it's actualy always a 32-bit floating point value in the range of [0.0, 1.0] or [-1.0,+1.0] for 
# uni and bipolar respectively, which will then scale to whatever translation the receiving device fancies, so consult the manuals for
# the relevant devices if there is something surprising you about the results.
cv1Low = 150
cv1High = 255
cv2Low = 1
cv2High = 80

accentProb = 25 # As a % of how likely the accent is to be present
maxSteps = 128 # 


#Mutation Params - done as globals, see no need to mess around with this too much.
globalRate = 7 # Default is 7=1/16th. 
# Full rate capability is: 0=1/1, 1=1/2, 2=1/4, 3=3/16, 4=1/4T, 5=1/8, 6=1/8T, 7=1/16, 8=1/16T, 9=1/32, 10=1/64)
globalShuffle = 0 # 0=off, 128=Global, 1-127 is scaled from 50-75%
globalLength = 32 
globalGate = 0.500 # This one is a float. If you want to customise, save the setting to a patch and inspect (or use Recon, of course).


# V1.0 Advanced options
cv1On = True
cv2On = True
accentsOn = True
tieFlagsOn = True
doubledPattern = True
progressionDoubling = False


#
# Leave everything below this line untouched.
#

def main():
    Plums = str(datetime.datetime.now())
    Year = Plums[0:4]
    Month = Plums[5:7]
    Day = Plums[8:10]
    Hour = Plums[11:13]
    Mins = Plums[14:16]
    Secs = Plums[17:19]
    directory = "Pat_The_Mutt_" + Year + Month + Day + Hour + Mins + Secs
    path = "./" + directory
    os.mkdir(path)
    i = 1
    for i in range(1,numOfPatches +1):
        textFilePath = path + "/Pat_Doggie_" + str(i) + ".repatch"
        i = i + 1
        with open(textFilePath, 'a') as f:
            f.write("<?xml version=\"1.0\"?>\n")
            f.write("<JukeboxPatch version=\"2.0\"  deviceProductID=\"se.propellerheads.Mutator\"  deviceVersion=\"1.0.4\" >\n")
            f.write("   <DeviceNameInEnglish>\n")
            f.write("       Pattern Mutator\n")
            f.write("   </DeviceNameInEnglish>\n")
            f.write("   <Properties>\n")
            f.write("       <Object name=\"custom_properties\" >\n")
            # Start of properties
            f.write("           <Value property=\"transpose_method\"  type=\"number\" >\n")
            f.write("               1\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"midi_velocity\"  type=\"boolean\" >\n")
            f.write("               false\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"manual_transpose_amount\"  type=\"number\" >\n")
            f.write("               24\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"playback_mode\"  type=\"number\" >\n")
            f.write("               2\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"selected_view\"  type=\"number\" >\n")
            f.write("               0\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"rate_1\"  type=\"number\" >\n")
            f.write("               " + str(globalRate) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"rate_2\"  type=\"number\" >\n")
            f.write("               " + str(globalRate) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"rate_3\"  type=\"number\" >\n")
            f.write("               " + str(globalRate) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"rate_4\"  type=\"number\" >\n")
            f.write("               " + str(globalRate) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"rate_5\"  type=\"number\" >\n")
            f.write("               " + str(globalRate) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"rate_6\"  type=\"number\" >\n")
            f.write("               " + str(globalRate) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"rate_7\"  type=\"number\" >\n")
            f.write("               " + str(globalRate) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"rate_8\"  type=\"number\" >\n")
            f.write("               " + str(globalRate) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"mutation_rate\"  type=\"number\" >\n")
            f.write("               " + str(globalRate) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"shuffle_1\"  type=\"number\" >\n")
            f.write("               " + str(globalShuffle) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"shuffle_2\"  type=\"number\" >\n")
            f.write("               " + str(globalShuffle) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"shuffle_3\"  type=\"number\" >\n")
            f.write("               " + str(globalShuffle) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"shuffle_4\"  type=\"number\" >\n")
            f.write("               " + str(globalShuffle) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"shuffle_5\"  type=\"number\" >\n")
            f.write("               " + str(globalShuffle) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"shuffle_6\"  type=\"number\" >\n")
            f.write("               " + str(globalShuffle) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"shuffle_7\"  type=\"number\" >\n")
            f.write("               " + str(globalShuffle) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"shuffle_8\"  type=\"number\" >\n")
            f.write("               " + str(globalShuffle) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"mutation_shuffle\"  type=\"number\" >\n")
            f.write("               " + str(globalShuffle) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"pattern_length_1\"  type=\"number\" >\n")
            f.write("               " + str(globalLength) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"pattern_length_2\"  type=\"number\" >\n")
            f.write("               " + str(globalLength) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"pattern_length_3\"  type=\"number\" >\n")
            f.write("               " + str(globalLength) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"pattern_length_4\"  type=\"number\" >\n")
            f.write("               " + str(globalLength) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"pattern_length_5\"  type=\"number\" >\n")
            f.write("               " + str(globalLength) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"pattern_length_6\"  type=\"number\" >\n")
            f.write("               " + str(globalLength) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"pattern_length_7\"  type=\"number\" >\n")
            f.write("               " + str(globalLength) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"pattern_length_8\"  type=\"number\" >\n")
            f.write("               " + str(globalLength) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"mutation_length\"  type=\"number\" >\n")
            f.write("               " + str(globalLength) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"gate_length_1\"  type=\"number\" >\n")
            f.write("               " + str(globalGate) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"gate_length_2\"  type=\"number\" >\n")
            f.write("               " + str(globalGate) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"gate_length_3\"  type=\"number\" >\n")
            f.write("               " + str(globalGate) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"gate_length_4\"  type=\"number\" >\n")
            f.write("               " + str(globalGate) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"gate_length_5\"  type=\"number\" >\n")
            f.write("               " + str(globalGate) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"gate_length_6\"  type=\"number\" >\n")
            f.write("               " + str(globalGate) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"gate_length_7\"  type=\"number\" >\n")
            f.write("               " + str(globalGate) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"gate_length_8\"  type=\"number\" >\n")
            f.write("               " + str(globalGate) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"mutation_gate_length\"  type=\"number\" >\n")
            f.write("               " + str(globalGate) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"slide\"  type=\"number\" >\n")
            f.write("               120\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"requested_buffer_length\"  type=\"number\" >\n")
            f.write("               32\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"keyboard_trigger\"  type=\"boolean\" >\n")
            f.write("               true\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"retrig\"  type=\"boolean\" >\n")
            f.write("               true\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"mutation_note_data\"  type=\"string\" />\n")
            f.write("           <Value property=\"mutation_start_note\"  type=\"number\" >\n")
            f.write("               128\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"octave_pitch_amount\"  type=\"number\" >\n")
            f.write("               0\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"accent_velocity_amount\"  type=\"number\" >\n")
            f.write("               0\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"juggle_amount\"  type=\"number\" >\n")
            f.write("               0\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"rhythm_amount\"  type=\"number\" >\n")
            f.write("               0\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"note_length_amount\"  type=\"number\" >\n")
            f.write("               0\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"mutation_cv_lane_1\"  type=\"string\" />\n")
            f.write("           <Value property=\"mutation_cv_lane_2\"  type=\"string\" />\n")
            f.write("           <Value property=\"mutation_accent_lane\"  type=\"string\" />\n")
            f.write("           <Value property=\"cv1_polarity\"  type=\"number\" >\n")
            f.write("               0\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"cv2_polarity\"  type=\"number\" >\n")
            f.write("               0\n")
            f.write("           </Value>\n")

            # Space reserved for the Pattern 1 Strings
            preWrites = generatePattern()
            f.write("           <Value property=\"pattern_data_1\"  type=\"string\" >\n")
            f.write("")
            f.write(str(preWrites[0]) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"start_note_1\"  type=\"number\" >\n")
            f.write("               " + str(preWrites[1]) + "\n")
            firstStep = 129 # Safe to reset this now
            firstNote = 0
            f.write("           </Value>\n")
            f.write("           <Value property=\"cv_lane_1_1\"  type=\"string\" >\n")
            f.write("")
            f.write(str(preWrites[2]) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"cv_lane_2_1\"  type=\"string\" >\n")
            f.write("")
            f.write(str(preWrites[3]) + "\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"accent_lane_1\"  type=\"string\" >\n")
            f.write("")
            f.write(str(preWrites[4]) + "\n")
            f.write("           </Value>\n")

            # Empty Patterns
            f.write("           <Value property=\"pattern_data_2\"  type=\"string\" />\n")
            f.write("           <Value property=\"start_note_2\"  type=\"number\" >\n")
            f.write("               128\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"cv_lane_1_2\"  type=\"string\" />\n")
            f.write("           <Value property=\"cv_lane_2_2\"  type=\"string\" />\n")
            f.write("           <Value property=\"accent_lane_2\"  type=\"string\" />\n")

            f.write("           <Value property=\"pattern_data_3\"  type=\"string\" />\n")
            f.write("           <Value property=\"start_note_3\"  type=\"number\" >\n")
            f.write("               128\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"cv_lane_1_3\"  type=\"string\" />\n")
            f.write("           <Value property=\"cv_lane_2_3\"  type=\"string\" />\n")
            f.write("           <Value property=\"accent_lane_3\"  type=\"string\" />\n")

            f.write("           <Value property=\"pattern_data_4\"  type=\"string\" />\n")
            f.write("           <Value property=\"start_note_4\"  type=\"number\" >\n")
            f.write("               128\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"cv_lane_1_4\"  type=\"string\" />\n")
            f.write("           <Value property=\"cv_lane_2_4\"  type=\"string\" />\n")
            f.write("           <Value property=\"accent_lane_4\"  type=\"string\" />\n")

            f.write("           <Value property=\"pattern_data_5\"  type=\"string\" />\n")
            f.write("           <Value property=\"start_note_5\"  type=\"number\" >\n")
            f.write("               128\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"cv_lane_1_5\"  type=\"string\" />\n")
            f.write("           <Value property=\"cv_lane_2_5\"  type=\"string\" />\n")
            f.write("           <Value property=\"accent_lane_5\"  type=\"string\" />\n")

            f.write("           <Value property=\"pattern_data_6\"  type=\"string\" />\n")
            f.write("           <Value property=\"start_note_6\"  type=\"number\" >\n")
            f.write("               128\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"cv_lane_1_6\"  type=\"string\" />\n")
            f.write("           <Value property=\"cv_lane_2_6\"  type=\"string\" />\n")
            f.write("           <Value property=\"accent_lane_6\"  type=\"string\" />\n")

            f.write("           <Value property=\"pattern_data_7\"  type=\"string\" />\n")
            f.write("           <Value property=\"start_note_7\"  type=\"number\" >\n")
            f.write("               128\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"cv_lane_1_7\"  type=\"string\" />\n")
            f.write("           <Value property=\"cv_lane_2_7\"  type=\"string\" />\n")
            f.write("           <Value property=\"accent_lane_7\"  type=\"string\" />\n")

            f.write("           <Value property=\"pattern_data_8\"  type=\"string\" />\n")
            f.write("           <Value property=\"start_note_8\"  type=\"number\" >\n")
            f.write("               128\n")
            f.write("           </Value>\n")
            f.write("           <Value property=\"cv_lane_1_8\"  type=\"string\" />\n")
            f.write("           <Value property=\"cv_lane_2_8\"  type=\"string\" />\n")
            f.write("           <Value property=\"accent_lane_8\"  type=\"string\" />\n")

            # End properties
            f.write("       </Object>\n")
            f.write("       <Object name=\"transport\" />\n")
            f.write("   </Properties>\n")
            f.write("</JukeboxPatch>\n")
            f.close()

def generatePattern():
    # Outputs
    patternString = ""
    startStep = 128
    startNote = 128
    cv1String = ""
    cv2String = ""
    accentString = ""
    doubledPatternString = ""
    # Get prospective number of events
    numEvents = random.randint(minNumOfEvents, maxNumOfEvents)

    if doubledPattern == True:
        # Sort the string headers
        doubledNumEvents = numEvents*2
        # Add leading 0 to event count - Python issue
        numEventsString = str(hex(doubledNumEvents + 1 + 256))
        numEventsChars = "0" + numEventsString[2:5]
        # Add format number and number of events
        patternString = "02" + numEventsChars #+ numEvents
        patternStringPost = ""  
        counter = 0
        #build the string events
        for each in range(1, numEvents + 1):
            newEvent = buildEvent()
            if newEvent[1] < startStep:
                startNote = newEvent[2]

            if progressionDoubling == True:
            # Allows the patternString to continue generating after the halfway point, 
            # giving a kind of "pseudo harmony" result.
                patternString = patternString + newEvent[0]
                doubledPatternString = doubledPatternString + newEvent[3]
            else:
            # if the event start number is lower than the halfway point, it gets in.
            # Otherwise it gets filtered.
            # The string header needs to get rewritten if this is the case as it will have less
            # events than is written elsewhere.                              
                eventStart = int.from_bytes(bytes.fromhex(newEvent[0][0:2]), byteorder='big')
                if eventStart <= int(globalLength/2): 
                    #print("I can fucking see me")
                    patternStringPost = patternStringPost + newEvent[0]
                    doubledPatternString = doubledPatternString + newEvent[3]
                    counter = counter + 2
                    #doubledPatternString = doubledPatternString + newEvent[3]



        # Export the now divergent string types
        if progressionDoubling == False:
            numEventsString = str(hex(counter + 1 + 256))
            numEventsChars = "0" + numEventsString[2:5]
            patternString = "02" + numEventsChars + patternStringPost + doubledPatternString
        else:
            patternString = patternString + doubledPatternString

        
        
        if cv1On == True:
            cv1String = "01"
            for each in range(0 , globalLength):
                cvEvent = str(hex(random.randint(cv1Low, cv1High)))[2:4]
                if len(cvEvent) < 2:
                    cvEvent = "0" + cvEvent
                cv1String = cv1String + cvEvent
            for each in range(0 , maxSteps - globalLength):
                cv1String = cv1String + "01"

        if cv2On == True:
            cv2String = "01"   
            for each in range(0 , globalLength):
                cvEvent = str(hex(random.randint(cv2Low, cv2High)))[2:4]
                if len(cvEvent) < 2:
                    cvEvent = "0" + cvEvent
                cv2String = cv2String + cvEvent
            for each in range(0 , maxSteps - globalLength):
                cv2String = cv2String + "01"

        if accentsOn == True:
            accentString = "01"    
            for each in range(0 , globalLength):
                if random.randint(1, 100) > accentProb:
                    accentString = accentString + "FF"
            for each in range(0 , maxSteps - globalLength):
                accentString = accentString + "01" 
    else:
        # Add leading 0 to eventCount - Python issue
        numEventsString = str(hex(numEvents + 1 + 256))
        numEventsChars = "0" + numEventsString[2:5]
        # Add format number and number of events
        patternString = "02" + numEventsChars
        for each in range(1, numEvents + 1):
            newEvent = buildEvent()
            patternString = patternString + newEvent[0]
            if newEvent[1] < startStep:
                startNote = newEvent[2]

        if cv1On == True:
            cv1String = "01"
            for each in range(0 , globalLength):
                cvEvent = str(hex(random.randint(cv1Low, cv1High)))[2:4]
                if len(cvEvent) < 2:
                    cvEvent = "0" + cvEvent
                cv1String = cv1String + cvEvent
            for each in range(0 , maxSteps - globalLength):
                cv1String = cv1String + "01"

        if cv2On == True:
            cv2String = "01"   
            for each in range(0 , globalLength):
                cvEvent = str(hex(random.randint(cv2Low, cv2High)))[2:4]
                if len(cvEvent) < 2:
                    cvEvent = "0" + cvEvent
                cv2String = cv2String + cvEvent
            for each in range(0 , maxSteps - globalLength):
                cv2String = cv2String + "01"

        if accentsOn == True:
            accentString = "01"    
            for each in range(0 , globalLength):
                if random.randint(1, 100) > accentProb:
                    accentString = accentString + "01"
                else:
                    accentString = accentString + "FF"
            for each in range(0 , maxSteps - globalLength):
                accentString = accentString + "01" 


    return patternString, startNote, cv1String, cv2String, accentString

def buildEvent():

#01 <!-- Note start (in steps, 1 - 128); in this case it's 1 = hex01 -->
    noteStartInt = random.randint(noteStartMin, noteStartMax)
    doubledNoteStartInt = noteStartInt + int(globalLength/2)
    noteStart = str(hex(noteStartInt))[2:4]
    if len(noteStart) < 2:
        noteStart = "0" + noteStart
    #f.write("Finished the Note Start")
    doubledNoteStart = str(hex(doubledNoteStartInt))[2:4]
    if len(doubledNoteStart) < 2:
        doubledNoteStart = "0" + doubledNoteStart

#3D <!-- Note number (1 - 128, MIDI note number offset by 1); in this case it's 60 (+1) = 61 = hex3D -->
    noteNumberInt = random.randint(noteRangeLow, noteRangeHigh)
    noteNumber = str(hex(noteNumberInt))[2:4]
    if len(noteNumber) < 2:
        noteNumber = "0" + noteNumber
    #f.write("Finished the Note Number")    

#48 <!-- Note velocity (1 - 128, i e offset by 1); in this case it's 71 (+1) = 72 = hex48 -->
    noteVelocity = str(hex(random.randint(noteVelLow, noteVelHigh)))[2:4]
    if len(noteVelocity) < 2:
        noteVelocity = "0" + noteVelocity

#02 <!-- Note length (in steps, offset by 1); in this case it's 1 (+1) = 2 = hex02 -->
    noteLength = str(hex(random.randint(noteLengthLow, noteLengthHigh)))[2:4]
    if len(noteLength) < 2:
        noteLength = "0" + noteLength

#93 <!-- Note flags; in this case it's 146 {128, 16(tied flag), +2(Octave offset; actually is -1 because 3 is the "0")} (+1) = 147 = hex93 (see NOTE 1 below) -->
#<!-- NOTE 1: The note flags value is 128 + (32 if the note is muted) + (16 if the note is tied) + the octave offset (0 - 6 corresponds to -3 - +3. the normal value is 3). 
#For a plain unmodified note, the flag value is 128 + 3 = 131, or "83" in hex. The per-note mute and octave offset are not normally used by Mutator, they were inherited 
#from Sequences. But the playback engine supports them. We don't need to know the C++ implementation details, but it's a safe bet this was either a well rounded switch or bit-shifting. -->
    if tieFlagsOn == True:
        tieProbability = random.randint(0, 100)
        if tieProbability > 70:
            noteFlags = "94"
        else: 
            noteFlags = "84"

#39 <!-- Length scale factor (in 1/254ths, offset by 1); in this case it's 56 (+1) = 57 = hex39 (see NOTE 2 below) -->
# Will need to test this a bit to find out what it actually does, before making any refinements.
    noteLSF = str(hex(random.randint(noteLSFLow, noteLSFHigh)))[2:4]
    if len(noteLSF) < 2:
        noteLSF = "0" + noteLSF

    event = noteStart + noteNumber + noteVelocity + noteLength + noteFlags + noteLSF
    doubledEvent = doubledNoteStart + noteNumber + noteVelocity + noteLength + noteFlags + noteLSF

    return event, noteStartInt, noteNumberInt, doubledEvent


if __name__ == "__main__":
    main()





